@@ -65,6 +65,7 @@ from urllib.parse import urlsplit
 from . import credentials
 from . import OscConfigParser
 from . import oscerr
+from .output import tty
 from .util import xdg
 from .util.helper import raw_input
 from .util.models import *
@@ -119,6 +120,12 @@ class Password(collections.UserString):
             return f"{self.__str__():{format_spec}}"
         return super().__format__(format_spec)
 
+    def encode(self, *args, **kwargs):
+        if sys.version_info < (3, 8):
+            # avoid returning the Password object on python < 3.8
+            return str(self).encode(*args, **kwargs)
+        return super().encode(*args, **kwargs)
+
 
 HttpHeader = NewType("HttpHeader", Tuple[str, str])
 
@@ -126,7 +133,9 @@ HttpHeader = NewType("HttpHeader", Tuple[str, str])
 class OscOptions(BaseModel):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
-        self.extra_fields = {}
+        self._allow_new_attributes = True
+        self._extra_fields = {}
+        self._allow_new_attributes = False
 
     # compat function with the config dict
     def _get_field_name(self, name):
@@ -145,7 +154,7 @@ class OscOptions(BaseModel):
         field_name = self._get_field_name(name)
 
         if field_name is None and not hasattr(self, name):
-            return self.extra_fields[name]
+            return self._extra_fields[name]
 
         field_name = field_name or name
         try:
@@ -158,7 +167,7 @@ class OscOptions(BaseModel):
         field_name = self._get_field_name(name)
 
         if field_name is None and not hasattr(self, name):
-            self.extra_fields[name] = value
+            self._extra_fields[name] = value
             return
 
         field_name = field_name or name
@@ -294,10 +303,20 @@ class HostOptions(OscOptions):
             """
             A pointer to public SSH key that corresponds with a private SSH used for authentication:
 
+             - keep empty for auto detection
              - path to the public SSH key
              - public SSH key filename (must be placed in ~/.ssh)
+             - fingerprint of a SSH key (2nd column of ``ssh-add -l``)
 
             NOTE: The private key may not be available on disk because it could be in a GPG keyring, on YubiKey or forwarded through SSH agent.
+
+            TIP: To give osc a hint which ssh key from the agent to use during auto detection,
+            append ``obs=<apiurl-hostname>`` to the **private** key's comment.
+            This will also work nicely during SSH agent forwarding, because the comments get forwarded too.
+
+             - To edit the key, run: ``ssh-keygen -c -f ~/.ssh/<private-key>``
+             - To query the key, run: ``ssh-keygen -y -f ~/.ssh/<private-key>``
+             - Example comment: ``<username@host> obs=api.example.com obs=api-test.example.com``
             """
         ),
     )  # type: ignore[assignment]
@@ -551,13 +570,25 @@ class Options(OscOptions):
         section=True,
     )  # type: ignore[assignment]
 
+    quiet: bool = Field(
+        default=False,
+        description=textwrap.dedent(
+            """
+            Reduce amount of printed information to bare minimum.
+            If enabled, automatically sets ``verbose`` to ``False``.
+            """
+        ),
+    )  # type: ignore[assignment]
+
     verbose: bool = Field(
         default=False,
         description=textwrap.dedent(
             """
             Increase amount of printed information to stdout.
+            Automatically set to ``False`` when ``quiet`` is enabled.
             """
         ),
+        get_callback=lambda conf, value: False if conf.quiet else value,
     )  # type: ignore[assignment]
 
     debug: bool = Field(
@@ -574,8 +605,10 @@ class Options(OscOptions):
         description=textwrap.dedent(
             """
             Print HTTP traffic to stderr.
+            Automatically set to ``True`` when``http_full_debug`` is enabled.
             """
         ),
+        get_callback=lambda conf, value: True if conf.http_full_debug else value,
     )  # type: ignore[assignment]
 
     http_full_debug: bool = Field(
@@ -583,6 +616,7 @@ class Options(OscOptions):
         description=textwrap.dedent(
             """
             [CAUTION!] Print HTTP traffic incl. authentication data to stderr.
+            If enabled, automatically sets ``http_debug`` to ``True``.
             """
         ),
     )  # type: ignore[assignment]
@@ -980,7 +1014,7 @@ class Options(OscOptions):
             The number of parallel processes during the build.
             Defaults to the number of available CPU threads.
 
-            Passed as ``--jobs`` to the build tool.
+            If the value is greater than ``0`` then it is passed as ``--jobs`` to the build tool.
             """
         ),
         ini_key="build-jobs",
@@ -1324,6 +1358,10 @@ def _model_to_rst(cls, title=None, description=None, sections=None, output_file=
         if field.default is None:
             return None
 
+        if field.default_is_lazy:
+            # lazy default may return different results under different circumstances -> return nothing
+            return None
+
         ini_type = field.extra.get("ini_type", None)
         if ini_type:
             return None
@@ -1377,10 +1415,10 @@ def _model_to_rst(cls, title=None, description=None, sections=None, output_file=
 
         ini_key = extra.get("ini_key", name)
 
-        x = bold(ini_key) + " : " + get_type(name, field)
+        x = f"{bold(ini_key)} : {get_type(name, field)}"
         default = get_default(name, field)
         if default:
-            x += " = " + italic(default)
+            x += f" = {italic(default)}"
         result.append(x)
         result.append("")
         desc = extra.get("ini_description", None) or field.description or ""
@@ -1493,13 +1531,13 @@ def parse_apisrv_url(scheme, apisrv):
     elif scheme is not None:
         url = scheme + apisrv
     else:
-        url = "https://" + apisrv
+        url = f"https://{apisrv}"
     scheme, url, path = urlsplit(url)[0:3]
     return scheme, url, path.rstrip('/')
 
 
 def urljoin(scheme, apisrv, path=''):
-    return '://'.join([scheme, apisrv]) + path
+    return f"{scheme}://{apisrv}" + path
 
 
 def is_known_apiurl(url):
@@ -1535,7 +1573,7 @@ def get_apiurl_api_host_options(apiurl):
     apiurl = sanitize_apiurl(apiurl)
     if is_known_apiurl(apiurl):
         return config['api_host_options'][apiurl]
-    raise oscerr.ConfigMissingApiurl('missing credentials for apiurl: \'%s\'' % apiurl,
+    raise oscerr.ConfigMissingApiurl(f'missing credentials for apiurl: \'{apiurl}\'',
                                      '', apiurl)
 
 
@@ -1578,14 +1616,14 @@ def get_configParser(conffile=None, force_read=False):
 
 def write_config(fname, cp):
     """write new configfile in a safe way"""
-    if os.path.exists(fname) and not os.path.isfile(fname):
-        # only write to a regular file
-        return
-
     # config file is behind a symlink
     # resolve the symlink and continue writing the config as usual
     if os.path.islink(fname):
-        fname = os.readlink(fname)
+        fname = os.path.realpath(fname)
+
+    if os.path.exists(fname) and not os.path.isfile(fname):
+        # only write to a regular file
+        return
 
     # create directories to the config file (if they don't exist already)
     fdir = os.path.dirname(fname)
@@ -1596,14 +1634,14 @@ def write_config(fname, cp):
             if e.errno != errno.EEXIST:
                 raise
 
-    with open(fname + '.new', 'w') as f:
+    with open(f"{fname}.new", 'w') as f:
         cp.write(f, comments=True)
     try:
-        os.rename(fname + '.new', fname)
+        os.rename(f"{fname}.new", fname)
         os.chmod(fname, 0o600)
     except:
-        if os.path.exists(fname + '.new'):
-            os.unlink(fname + '.new')
+        if os.path.exists(f"{fname}.new"):
+            os.unlink(f"{fname}.new")
         raise
 
 
@@ -1636,10 +1674,10 @@ def config_set_option(section, opt, val=None, delete=False, update=True, creds_m
 
     section = sections.get(section.rstrip('/'), section)
     if section not in cp.sections():
-        raise oscerr.ConfigError('unknown section \'%s\'' % section, config['conffile'])
+        raise oscerr.ConfigError(f'unknown section \'{section}\'', config['conffile'])
     if section == 'general' and opt not in general_opts or \
        section != 'general' and opt not in api_host_options:
-        raise oscerr.ConfigError('unknown config option \'%s\'' % opt, config['conffile'])
+        raise oscerr.ConfigError(f'unknown config option \'{opt}\'', config['conffile'])
 
     if not val and not delete and opt == 'pass' and creds_mgr_descr is not None:
         # change password store
@@ -1716,7 +1754,7 @@ def write_initial_config(conffile, entries, custom_template='', creds_mgr_descri
     config.update(entries)
     sio = StringIO(conf_template.strip() % config)
     cp = OscConfigParser.OscConfigParser()
-    cp.readfp(sio)
+    cp.read_file(sio)
     cp.set(config['apiurl'], 'user', config['user'])
     if creds_mgr_descriptor:
         creds_mgr = creds_mgr_descriptor.create(cp)
@@ -1752,7 +1790,7 @@ def _get_credentials_manager(url, cp):
     if cp.has_option(url, credentials.AbstractCredentialsManager.config_entry):
         creds_mgr = credentials.create_credentials_manager(url, cp)
         if creds_mgr is None:
-            msg = 'Unable to instantiate creds mgr (section: %s)' % url
+            msg = f'Unable to instantiate creds mgr (section: {url})'
             conffile = get_configParser.conffile
             raise oscerr.ConfigMissingCredentialsError(msg, conffile, url)
         return creds_mgr
@@ -1770,6 +1808,7 @@ def get_config(override_conffile=None,
                override_http_full_debug=None,
                override_traceback=None,
                override_post_mortem=None,
+               override_quiet=None,
                override_no_keyring=None,
                override_verbose=None,
                overrides=None
@@ -1778,8 +1817,8 @@ def get_config(override_conffile=None,
     Configure osc.
 
     The configuration options are loaded with the following priority:
-        1. environment variables: OSC_<uppercase_option>
-        2. override arguments provided to get_config()
+        1. environment variables: ``OSC_<uppercase_option>`` or ``OSC_<uppercase_host_alias>_<uppercase_host_option>``
+        2. override arguments provided to ``get_config()``
         3. oscrc config file
     """
 
@@ -1798,7 +1837,6 @@ def get_config(override_conffile=None,
         overrides["http_debug"] = override_http_debug
 
     if override_http_full_debug is not None:
-        overrides["http_debug"] = override_http_full_debug or overrides["http_debug"]
         overrides["http_full_debug"] = override_http_full_debug
 
     if override_traceback is not None:
@@ -1810,6 +1848,9 @@ def get_config(override_conffile=None,
     if override_no_keyring is not None:
         overrides["use_keyring"] = not override_no_keyring
 
+    if override_quiet is not None:
+        overrides["quiet"] = override_quiet
+
     if override_verbose is not None:
         overrides["verbose"] = override_verbose
 
@@ -1818,29 +1859,34 @@ def get_config(override_conffile=None,
     else:
         conffile = identify_conf()
 
-    conffile = os.path.expanduser(conffile)
-    if not os.path.exists(conffile):
-        raise oscerr.NoConfigfile(conffile, account_not_configured_text % conffile)
+    if conffile in ["", "/dev/null"]:
+        cp = OscConfigParser.OscConfigParser()
+        cp.add_section("general")
+    else:
+        conffile = os.path.expanduser(conffile)
+        if not os.path.exists(conffile):
+            raise oscerr.NoConfigfile(conffile, account_not_configured_text % conffile)
 
-    # make sure oscrc is not world readable, it may contain a password
-    conffile_stat = os.stat(conffile)
-    if conffile_stat.st_mode != 0o600:
-        try:
-            os.chmod(conffile, 0o600)
-        except OSError as e:
-            if e.errno in (errno.EROFS, errno.EPERM):
-                print(f"Warning: Configuration file '{conffile}' may have insecure file permissions.")
-            else:
-                raise e
+        # make sure oscrc is not world readable, it may contain a password
+        conffile_stat = os.stat(conffile)
+        # applying 0o7777 mask because we want to ignore the file type bits
+        if conffile_stat.st_mode & 0o7777 != 0o600:
+            try:
+                os.chmod(conffile, 0o600)
+            except OSError as e:
+                if e.errno in (errno.EROFS, errno.EPERM):
+                    print(f"Warning: Configuration file '{conffile}' may have insecure file permissions.", file=sys.stderr)
+                else:
+                    raise e
 
-    cp = get_configParser(conffile)
+        cp = get_configParser(conffile)
 
-    if not cp.has_section('general'):
-        # FIXME: it might be sufficient to just assume defaults?
-        msg = config_incomplete_text % conffile
-        defaults = Options().dict()
-        msg += new_conf_template % defaults
-        raise oscerr.ConfigError(msg, conffile)
+        if not cp.has_section("general"):
+            # FIXME: it might be sufficient to just assume defaults?
+            msg = config_incomplete_text % conffile
+            defaults = Options().dict()
+            msg += new_conf_template % defaults
+            raise oscerr.ConfigError(msg, conffile)
 
     global config
 
@@ -1862,29 +1908,44 @@ def get_config(override_conffile=None,
     urls = [i for i in cp.sections() if i != "general"]
     for url in urls:
         apiurl = sanitize_apiurl(url)
-        username = cp[url].get("user", None)
-        if username is None:
-            raise oscerr.ConfigMissingCredentialsError(f"No user found in section {url}", conffile, url)
-
+        # the username will be overwritten later while reading actual config values
+        username = cp[url].get("user", "")
         host_options = HostOptions(apiurl=apiurl, username=username, _parent=config)
+
         known_ini_keys = set()
         for name, field in host_options.__fields__.items():
+            # the following code relies on interating through fields in a given order: aliases, username, credentials_mgr_class, password
+
             ini_key = field.extra.get("ini_key", name)
             known_ini_keys.add(ini_key)
 
-            if name == "password":
-                # we need to handle the password first because it may be stored in a keyring instead of a config file
-                creds_mgr = _get_credentials_manager(url, cp)
-                value = creds_mgr.get_password(url, host_options.username, defer=True, apiurl=host_options.apiurl)
-                if value is None:
-                    raise oscerr.ConfigMissingCredentialsError("No password found in section {url}", conffile, url)
-                value = Password(value)
+            # iterate through aliases and store the value of the the first env that matches OSC_HOST_{ALIAS}_{NAME}
+            env_value = None
+            for alias in host_options.aliases:
+                alias = alias.replace("-", "_")
+                env_key = f"OSC_HOST_{alias.upper()}_{name.upper()}"
+                env_value = os.environ.get(env_key, None)
+                if env_value is not None:
+                    break
+
+            if env_value is not None:
+                value = env_value
             elif ini_key in cp[url]:
                 value = cp[url][ini_key]
             else:
-                continue
+                value = None
 
-            host_options.set_value_from_string(name, value)
+            if name == "credentials_mgr_class":
+                # HACK: inject credentials_mgr_class back in case we have specified it from env to have it available for reading password
+                if value:
+                    cp[url][credentials.AbstractCredentialsManager.config_entry] = value
+            elif name == "password":
+                creds_mgr = _get_credentials_manager(url, cp)
+                if env_value is None:
+                    value = creds_mgr.get_password(url, host_options.username, defer=True, apiurl=host_options.apiurl)
+
+            if value is not None:
+                host_options.set_value_from_string(name, value)
 
         for key, value in cp[url].items():
             if key.startswith("_"):
@@ -1914,6 +1975,9 @@ def get_config(override_conffile=None,
         # priority: env, overrides, config
         if env_key in os.environ:
             value = os.environ[env_key]
+            # remove any matching records from overrides because they are checked for emptiness later
+            overrides.pop(name, None)
+            overrides.pop(ini_key, None)
         elif name in overrides:
             value = overrides.pop(name)
         elif ini_key in overrides:
@@ -1935,6 +1999,45 @@ def get_config(override_conffile=None,
 
         config.set_value_from_string(name, value)
 
+    # BEGIN: override credentials for the default apiurl
+
+    # OSC_APIURL is handled already because it's a regular field
+    env_username = os.environ.get("OSC_USERNAME", "")
+    env_credentials_mgr_class = os.environ.get("OSC_CREDENTIALS_MGR_CLASS", None)
+    env_password = os.environ.get("OSC_PASSWORD", None)
+
+    if config.apiurl not in config.api_host_options:
+        host_options = HostOptions(apiurl=config.apiurl, username=env_username, _parent=config)
+        config.api_host_options[config.apiurl] = host_options
+        # HACK: inject section so we can add credentials_mgr_class later
+        cp.add_section(config.apiurl)
+
+    host_options = config.api_host_options[config.apiurl]
+    if env_username:
+        host_options.set_value_from_string("username", env_username)
+
+    if env_credentials_mgr_class:
+        host_options.set_value_from_string("credentials_mgr_class", env_credentials_mgr_class)
+        # HACK: inject credentials_mgr_class in case we have specified it from env to have it available for reading password
+        cp[config.apiurl]["credentials_mgr_class"] = env_credentials_mgr_class
+
+    if env_password:
+        password = Password(env_password)
+        host_options.password = password
+    elif env_credentials_mgr_class:
+        creds_mgr = _get_credentials_manager(config.apiurl, cp)
+        password = creds_mgr.get_password(config.apiurl, host_options.username, defer=True, apiurl=host_options.apiurl)
+        host_options.password = password
+
+    # END: override credentials for the default apiurl
+
+    for apiurl, host_options in config.api_host_options.items():
+        if not host_options.username:
+            raise oscerr.ConfigMissingCredentialsError(f"No user configured for apiurl {apiurl}", conffile, apiurl)
+
+        if host_options.password is None:
+            raise oscerr.ConfigMissingCredentialsError(f"No password configured for apiurl {apiurl}", conffile, apiurl)
+
     for key, value in cp["general"].items():
         if key.startswith("_"):
             continue
@@ -1967,13 +2070,17 @@ def identify_conf():
     # needed for compat reasons(users may have their oscrc still in ~
     if 'OSC_CONFIG' in os.environ:
         return os.environ.get('OSC_CONFIG')
-    if os.path.exists(os.path.expanduser('~/.oscrc')):
-        return '~/.oscrc'
 
-    if os.environ.get('XDG_CONFIG_HOME', '') != '':
-        conffile = os.environ.get('XDG_CONFIG_HOME') + '/osc/oscrc'
-    else:
-        conffile = '~/.config/osc/oscrc'
+    conffile = os.path.join(xdg.XDG_CONFIG_HOME, "osc", "oscrc")
+
+    if os.path.exists(os.path.expanduser("~/.oscrc")) or os.path.islink(os.path.expanduser("~/.oscrc")):
+        if "XDG_CONFIG_HOME" in os.environ:
+            print(f"{tty.colorize('WARNING', 'yellow,bold')}: Ignoring XDG_CONFIG_HOME env, loading an existing config from '~/.oscrc' instead", file=sys.stderr)
+            print("         To fix this, move the existing '~/.oscrc' to XDG location such as '~/.config/osc/oscrc'", file=sys.stderr)
+        elif os.path.exists(os.path.expanduser(conffile)):
+            print(f"{tty.colorize('WARNING', 'yellow,bold')}: Ignoring config '{conffile}' in XDG location, loading an existing config from ~/.oscrc instead", file=sys.stderr)
+            print("         To fix this, remove '~/.oscrc'", file=sys.stderr)
+        return '~/.oscrc'
 
     return conffile
 
