@@ -21,10 +21,10 @@ import urllib3.response
 import urllib3.util
 
 from . import __version__
-from . import _private
 from . import conf
 from . import oscerr
 from . import oscssl
+from . import output
 from .util.helper import decode_it
 
 
@@ -141,12 +141,12 @@ def http_request_wrap_file(func):
     Turn file path into a file object and close it automatically
     by using a context manager.
     """
-    def new_func(method, url, headers=None, data=None, file=None, retry_on_400: bool = True):
+    def new_func(method, url, headers=None, data=None, file=None):
         if file:
             with open(file, "rb") as f:
-                return func(method, url, headers, data, f, retry_on_400)
+                return func(method, url, headers, data, file=f)
         else:
-            return func(method, url, headers, data, file, retry_on_400)
+            return func(method, url, headers, data, file)
 
     new_func.__name__ = func.__name__
     new_func.__doc__ = func.__doc__
@@ -154,7 +154,7 @@ def http_request_wrap_file(func):
 
 
 @http_request_wrap_file
-def http_request(method: str, url: str, headers=None, data=None, file=None, retry_on_400: bool = True):
+def http_request(method: str, url: str, headers=None, data=None, file=None):
     """
     Send a HTTP request to a server.
 
@@ -174,7 +174,6 @@ def http_request(method: str, url: str, headers=None, data=None, file=None, retr
     :param headers: Dictionary of custom headers to send.
     :param data: Data to send in the request body (conflicts with `file`).
     :param file: Path to a file to send as data in the request body (conflicts with `data`).
-    :param retry_on_400: Whether to retry on receiving HTTP status code 400.
     """
 
     purl = urllib3.util.parse_url(url)
@@ -234,22 +233,15 @@ def http_request(method: str, url: str, headers=None, data=None, file=None, retr
         else:
             retries_kwargs = {"method_whitelist": None}
 
-
-        status_forcelist = (
-            500,  # Internal Server Error
-            502,  # Bad Gateway
-            503,  # Service Unavailable
-            504,  # Gateway Timeout
-        )
-        if retry_on_400:
-            status_forcelist = (
-                400,  # Bad Request; retry on 400: service in progress
-            ) + status_forcelist
-
         pool_kwargs["retries"] = urllib3.Retry(
             total=int(conf.config["http_retries"]),
             backoff_factor=2,
-            status_forcelist=status_forcelist,
+            status_forcelist=(
+                500,  # Internal Server Error
+                502,  # Bad Gateway
+                503,  # Service Unavailable
+                504,  # Gateway Timeout
+            ),
             # don't raise because we want an actual response rather than a MaxRetryError with "too many <status_code> error responses" message
             raise_on_status=False,
             **retries_kwargs,
@@ -488,7 +480,7 @@ class CookieJarAuthHandler(AuthHandlerBase):
             # doing expensive signature auth in multiple processes.
             # This usually happens when a user runs multiple osc instances
             # from the command-line in parallel.
-            self.cookiejar_lock_path = self.cookiejar_path + ".lock"
+            self.cookiejar_lock_path = f"{self.cookiejar_path}.lock"
         self.cookiejar_lock_fd = None
 
     @property
@@ -574,7 +566,21 @@ class SignatureAuthHandler(AuthHandlerBase):
     def __init__(self, apiurl, user, sshkey, basic_auth_password=None):
         super().__init__(apiurl)
         self.user = user
-        self.sshkey = sshkey
+        self.sshkey = None
+        self.sshkey_fingerprint = None
+
+        if sshkey and re.match("^[A-Z0-9]+:.*", sshkey):
+            # if it starts with a prefix such as 'SHA256:' then it's a fingerprint
+            self.sshkey_fingerprint = sshkey
+        else:
+            self.sshkey = sshkey
+
+        if self.sshkey:
+            # if only a file name is provided, prepend ~/.ssh
+            if "/" not in self.sshkey:
+                self.sshkey = os.path.join("~", ".ssh", self.sshkey)
+                self.sshkey = os.path.expanduser(self.sshkey)
+            output.print_msg(f"Using ssh key file configured in oscrc: {self.sshkey}", print_to="debug")
 
         self.ssh_keygen_path = shutil.which("ssh-keygen")
         self.ssh_add_path = shutil.which("ssh-add")
@@ -592,53 +598,82 @@ class SignatureAuthHandler(AuthHandlerBase):
         if not self.ssh_add_path:
             return []
         cmd = [self.ssh_add_path, '-L']
-        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8")
         stdout, _ = proc.communicate()
         if proc.returncode == 0 and stdout.strip():
             return stdout.strip().splitlines()
         else:
             return []
 
-    @staticmethod
-    def get_fingerprint(line):
-        parts = line.strip().split(b" ")
-        if len(parts) < 2:
-            raise ValueError(f"Unable to retrieve ssh key fingerprint from line: {line}")
-        return parts[1]
+    def list_ssh_agent_fingerprints(self):
+        if not self.ssh_add_path:
+            return []
+        cmd = [self.ssh_add_path, '-l']
+        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8")
+        stdout, _ = proc.communicate()
+        if proc.returncode == 0 and stdout.strip():
+            lines = stdout.strip().splitlines()
+            return [i.split(" ")[1] for i in lines]
+        else:
+            return []
 
     def guess_keyfile(self):
         # `ssh-keygen -Y sign` requires a file with a key which is not available during ssh agent forwarding
         # that's why we need to list ssh-agent's keys and store the first one into a temp file
         keys_in_agent = self.list_ssh_agent_keys()
         if keys_in_agent:
-            self.temp_pubkey = tempfile.NamedTemporaryFile()
+            selected_key = None
+
+            # use ssh key from ssh agent by the specified fingerprint
+            if self.sshkey_fingerprint:
+                fingerprints_in_agent = self.list_ssh_agent_fingerprints()
+                try:
+                    indx = fingerprints_in_agent.index(self.sshkey_fingerprint)
+                    selected_key = keys_in_agent[indx]
+                    output.print_msg(f"Using ssh key from ssh agent that matches fingerprint '{self.sshkey_fingerprint}': {selected_key}", print_to="debug")
+                except ValueError:
+                    pass
+
+            # use ssh key from ssh agent by key's comment obs=<hostname> matching the hostname of apiurl
+            if selected_key is None:
+                apiurl_hostname = urllib.parse.urlparse(self.apiurl).hostname
+                for key in keys_in_agent:
+                    comments = key.strip().split(" ")[2:]
+                    pattern = f"obs={apiurl_hostname}"
+                    if pattern in comments:
+                        selected_key = key
+                        output.print_msg(f"Using ssh key from ssh agent that has comment '{pattern}' which matches apiurl '{self.apiurl}': {selected_key}", print_to="debug")
+                        break
+
+            # use the first ssh key from ssh agent
+            if selected_key is None:
+                selected_key = keys_in_agent[0]
+                output.print_msg(f"Using the first ssh key from ssh agent (see `ssh-add -L`): {selected_key}", print_to="debug")
+
+            self.temp_pubkey = tempfile.NamedTemporaryFile(mode="w+")
             self.temp_pubkey.write(keys_in_agent[0])
             self.temp_pubkey.flush()
             return self.temp_pubkey.name
 
         sshdir = os.path.expanduser('~/.ssh')
         keyfiles = ('id_ed25519', 'id_ed25519_sk', 'id_rsa', 'id_ecdsa', 'id_ecdsa_sk', 'id_dsa')
+        output.print_msg(f"Searching ssh keys in '{sshdir}' in the following order: {', '.join(keyfiles)}", print_to="debug")
         for keyfile in keyfiles:
             keyfile_path = os.path.join(sshdir, keyfile)
             if os.path.isfile(keyfile_path):
+                output.print_msg(f"Using ssh key from file: {keyfile_path}", print_to="debug")
                 return keyfile_path
         return None
 
     def ssh_sign(self, data, namespace, keyfile=None):
-        try:
-            data = bytes(data, 'utf-8')
-        except:
-            pass
         if not keyfile:
             keyfile = self.guess_keyfile()
-        else:
-            if '/' not in keyfile:
-                keyfile = '~/.ssh/' + keyfile
-            keyfile = os.path.expanduser(keyfile)
-
+        if not keyfile:
+            raise oscerr.OscIOError(None, "No SSH key configured or auto-detected")
+        keyfile = os.path.expanduser(keyfile)
         cmd = [self.ssh_keygen_path, '-Y', 'sign', '-f', keyfile, '-n', namespace, '-q']
-        proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
-        stdout, _ = proc.communicate(data)
+        proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, encoding="utf-8")
+        signature, _ = proc.communicate(data)
 
         if self.temp_pubkey:
             self.temp_pubkey.close()
@@ -647,7 +682,6 @@ class SignatureAuthHandler(AuthHandlerBase):
         if proc.returncode:
             raise oscerr.OscIOError(None, 'ssh-keygen signature creation failed: %d' % proc.returncode)
 
-        signature = decode_it(stdout)
         match = re.match(r"\A-----BEGIN SSH SIGNATURE-----\n(.*)\n-----END SSH SIGNATURE-----", signature, re.S)
         if not match:
             raise oscerr.OscIOError(None, 'could not extract ssh signature')
@@ -668,7 +702,7 @@ class SignatureAuthHandler(AuthHandlerBase):
         auth = self.get_authorization(chal)
         if not auth:
             return False
-        auth_val = 'Signature %s' % auth
+        auth_val = f'Signature {auth}'
         req.add('Authorization', auth_val)
         return True
 
@@ -690,7 +724,7 @@ class SignatureAuthHandler(AuthHandlerBase):
             return False
 
         if not self.ssh_keygen_path:
-            _private.print_msg("Skipping signature auth because ssh-keygen is not available", print_to="debug")
+            output.print_msg("Skipping signature auth because ssh-keygen is not available", print_to="debug")
             return False
 
         if not self.sshkey_known():
